#include <sys/types.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/unistd.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <netdb.h>
#include <signal.h>
#include <time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <ctype.h>
#include <fcntl.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <string.h>
#include <time.h>

char *strptime(const char *s, const char *format, struct tm *tm);

static const char MYSQL_TXT[] = "/etc/mythtv/mysql.txt";
static char *mysql_hostname, *mysql_username, *mysql_passwd, *mysql_dbname;

static void inline
err_exit(const char *msg)
{
	int err = errno;	// save errno value
	perror(msg);		// output error message
	exit(err);		// exit with errno value
}

static int
format_time (char *outbuf, int nbytes, time_t t, char *fmt)
{
	struct tm	lt;

	localtime_r(&t, &lt);
	return strftime(outbuf, nbytes, fmt, &lt);
}

static void
mysql_get_info (void)
{
	char buf[256];
	FILE *fp = fopen(MYSQL_TXT, "r");

	if (!fp)
		err_exit(MYSQL_TXT);
	while (fgets(buf, sizeof(buf), fp)) {
		sscanf(buf, "DBHostName=%as", &mysql_hostname);
		sscanf(buf, "DBUserName=%as", &mysql_username);
		sscanf(buf, "DBName=%as", &mysql_dbname);
		sscanf(buf, "DBPassword=%as", &mysql_passwd);
	}
	fclose(fp);
	if (!mysql_hostname)
		mysql_username = "localhost";
	if (!mysql_username)
		mysql_username = "mythtv";
	if (!mysql_dbname)
		mysql_dbname   = "mythconverg";
	if (!mysql_passwd) {
		fprintf(stderr, "couldn't get mysql passwd\n");
		exit(ENODEV);
	}
}

static int
mysql_get_encoder_list (int *encoders)
{
	/*
	 * The backend crashes if we query a non-existant "remoteencoder" cardid.
	 * So, we need to query the mySQL database first for a full/current list of them.
	 */
	char cmd[1024];
	FILE *fp;
	int n = 0;
	char buf[1024];

	sprintf(cmd, "mysql -h '%s' -u '%s' '--password=%s' '%s'"
		" -e \"SELECT cardid FROM capturecard WHERE parentid='0' ORDER BY cardid;\"",
		mysql_hostname, mysql_username, mysql_passwd, mysql_dbname);
	fp = popen(cmd, "r");
	if (!fp) {
		perror("popen(mysql)");
		return 0;
	}
	while (fgets(buf, sizeof(buf), fp)) {
		if (1 == sscanf(buf, "%u", &encoders[n])) {
			//printf("found encoder %u\n", encoders[n]);
			++n;
		}
	}
	fclose(fp);
	return n;
}

static int
send_backend_request (int sd, const char *text)
{
	int text_len = strlen(text);
	int msg_len = 8 + text_len;
	int rc, err;
	char *msg = malloc(msg_len + 1);

	if (msg == NULL) {
		perror("malloc()");
		return -ENOMEM;
	}
	sprintf(msg, "%-8u%s", text_len, text);

	rc = write(sd, msg, msg_len);
	err = errno;
	free(msg);
	if (rc != msg_len) {
		if (rc == -1) {
			errno = err;
			perror("write(sd)");
			return err;
		} else {
			fprintf(stderr, "write(sd,%d): returned %d\n", msg_len, rc);
			return EIO;
		}
	}
	return 0;
}

static char *
get_backend_reply (int sd)
{
	int rc, len;
	char msg_len[9], *msg;

	rc = read(sd, msg_len, 8);
	if (rc < 8) {
		if (rc == -1) {
			perror("read(socket)");
			return NULL;
		}
		fprintf(stderr, "got short reply: %d bytes\n", rc);
		return NULL;
	}
	msg_len[8] = '\0';
	len = atoi(msg_len);
	if (len <= 0) {
		fprintf(stderr, "got short msg_len: %d bytes\n", len);
		return NULL;
	}
	msg = malloc(len + 1);
	if (!msg) {
		perror("malloc()");
		return NULL;
	}
	msg[len] = '\0';
	rc = read(sd, msg, len);
	if (rc != len) {
		if (rc == -1)
			perror("read(socket)");
		else
			fprintf(stderr, "msg_len(%d) != actual(%d)\n", len, rc);
		free(msg);
		return NULL;
	}
	return msg;
}

static char *
do_backend_command (void *vfd, int sd, char *command)
{
	char *reply;

	if (send_backend_request(sd, command))
		return NULL;
	reply = get_backend_reply(sd);
	return reply;
}

static char *
substring (char *reply, int field, char **internal_p)
{
	int i;
	char *next, *new, saved;
	static const char *magic = "[]:[]";

	//printf("substring(%p(0x%02x),%d,%p): ", reply, reply ? reply[0] : 0, field, internal_p); fflush(stdout);
	for (i = 0; i != field; ++i) {
		if (!*reply || !(next = strstr(reply, magic))) {
			//printf("NULL\n"); fflush(stdout);
			return NULL;
		}
		reply = next + strlen(magic);
	}

	next = strstr(reply, magic);
	if (!next)
		next = reply + strlen(reply);
	saved = *next;
	*next = '\0';
	new = malloc(strlen(reply)+1);
	if (new)
		strcpy(new, reply); 
	*next = saved;
	if (internal_p)
		*internal_p = reply;
	//printf("%p\n", new); fflush(stdout);
	return new;
}

static void
format_and_scroll (char *outbuf, const char *text, int tick)
{
	int i, t = 0, start = (tick < 15) ? 15 - tick : 0;
	static int blink = 0;

	if (tick > 15) {
		int tlen = strlen(text);
		t = tick - 15;
		if (t > tlen)
			t = tlen;
	}
	for (i = 0; i < 16; ++i)
		outbuf[i] = (i < start || !text[t]) ? ' ' : text[t++];
	format_time(outbuf+16, 17, time(NULL), " %H:%M   %d-%b ");
	if (text[0]) {
		blink = 0;
	} else if (blink++ == 2) {
		outbuf[19] = ' ';
		blink = 0;
	}
}

static unsigned long long
get_msecs (void)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (tv.tv_sec * 1000) + (tv.tv_usec / 1000);
}

static void
dump_reply_fields (char *reply, int min, int max)
{
	int i;

	for (i = min; i <= max; ++i) {
		char *f = substring(reply,i,NULL);
		if (f) {
			printf("field[%2u] = \"%s\"\n", i, f);
			free(f);
		}
	}
	putchar('\n');
}

static void
monitor_backend (void *vfd, int sd)
{
	int	i, encoders[32], nencoders;
	char	*reply, myhost[64], command[128];
	unsigned long max_hours = (7 * 24);
	time_t	last_recording_update = time(NULL);

	// Log-in to the backend:
	if (!(reply = do_backend_command(vfd, sd, "MYTH_PROTO_VERSION 31")))
		return;
	free(reply);

	// Get a list of recording devices from mySQL
	// for later use in generating queries to the backend:
	nencoders = mysql_get_encoder_list(encoders);

	// Register with the backend as a "Monitor" machine:
	gethostname(myhost, sizeof(myhost));
	sprintf(command, "ANN Monitor %s 0", myhost);
	if (!(reply = do_backend_command(vfd, sd, command)))
		return;
	free(reply);

	// Now just loop, querying the backend and showing the results:
	for (;;) {
		int is_recording, i;
		char *r, text[4096];
		int count = 0;
		time_t next_tstart;

		if (!(reply = do_backend_command(vfd, sd, "QUERY_ISRECORDING")))
			return;
		is_recording = (reply[0] != '0');
		free(reply);
	
		if (!(reply = do_backend_command(vfd, sd, "QUERY_GETALLPENDING")))
			return;

		// First pass: find earliest recording time from the list:
		next_tstart = 0;
		for (r = reply; *r && strlen(r) > (40 * strlen("[]:[]"));) {
			char *start, *next;
			time_t tstart;

			start = substring(r,13,NULL);
			if (!start)
				break;
			tstart = atoi(start);
			if (!next_tstart || tstart < next_tstart)
				next_tstart = tstart;
			free(start);
			next = substring(r,42,&r);
			if (!next)
				break;
			free(next);
		}
		if (!next_tstart) {
			strcpy(text, "(nothing scheduled)");
		} else {
			// Second pass: find all shows which match the earliest recording time:
			for (r = reply; *r && strlen(r) > (40 * strlen("[]:[]"));) {
				char *start, *next;
				time_t tstart;
	
				start = substring(r,13,NULL);
				if (!start)
					break;
				tstart = atoi(start);
				if (tstart == next_tstart) {
					char *title = substring(r,2,NULL);
					char *channel = substring(r,7,NULL);
					char *end = substring(r,14,NULL);
	
					if (title && channel && end) {
						time_t tend = atoi(end);
						char fstart[32], fend[32];
						format_time(fstart, sizeof(fstart), tstart, "%F %R");
						format_time(fend,   sizeof(fend),   tend,   "%R");
						count += snprintf(text+count, sizeof(text)-count,
							"Ch.%s: %s, %s-%s.  ", channel, title, fstart, fend);
					}
					if (title) free(title);
					if (channel) free(channel);
					if (end) free(end);
				}
				free(start);
				next = substring(r,42,&r);
				if (!next)
					break;
				free(next);
			}
		}
		free(reply);

		// Scroll the long string across the VFD:
		printf("%s\n", text);
		exit(0);
	}
}

int
main (int argc, char *argv[])
{
	int rc, sd;
	struct sockaddr_in saddr;
	void *vfd = NULL;
	char *myname = argv[0];

	mysql_get_info();

	for (;;) {
		// create a new socket object:
		sd = socket(PF_INET, SOCK_STREAM, 6 /* TCP */);
		if (sd == -1)
			err_exit("socket() failed");
	
		// establish a "name" (ipaddr:port) for the server:
		memset(&saddr, 0, sizeof(saddr));
		saddr.sin_family = AF_INET;
		saddr.sin_addr.s_addr = htonl(0x7f000001);// localhost: 127.0.0.1
		saddr.sin_port = htons(6543);		// mythbackend port number

		// (re-)connect to the server:
		rc = connect(sd, (struct sockaddr *)&saddr, sizeof(saddr));
		if (rc == -1) {
			if (errno != ECONNREFUSED && errno != ETIMEDOUT)
				err_exit("connect() failed");
			perror("connect()");
		} else {
			// handle the connection
			monitor_backend(vfd, sd);
		}
		close(sd);
	}
	return 0;
}
