#include <sys/types.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/unistd.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <netdb.h>
#include <signal.h>
#include <time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <ctype.h>
#include <fcntl.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <string.h>
#include <time.h>
#include "schema.h"

char *strptime(const char *s, const char *format, struct tm *tm);

static const char MYSQL_TXT[] = "/etc/mythtv/mysql.txt";
static char *mysql_hostname, *mysql_username, *mysql_passwd, *mysql_dbname;
static char backend_hostname[64];
static char *errtext;


// These are in vfd_usb.c:
//
void *vfd_open(void);
void vfd_close(void *);
int vfd_write(void *, char *);
int knob_read(void *);

static void inline
err_exit(const char *msg)
{
	int err = errno;
	syslog(LOG_ERR, "%s: %s\n", msg, strerror(err));
	exit(err);
}

// We trap common signals, to allow blanking the display
// before our program exits.  This way, the user can see
// that we're not actually updating the display anymore.
//
static volatile int shutdown_sig = 0;
static void
my_sighandler (int sig)
{
	shutdown_sig = sig;
}

// It's not safe to reenter vfd_write() from my_sighandler(),
// so instead we have the mainline thread call this function:
//
static void
kill_ourself (void *vfd)
{
	vfd_write(vfd, "");
	signal(shutdown_sig, SIG_DFL);
	raise(shutdown_sig);
	pause();
}

static int
format_time (char *outbuf, int nbytes, time_t t, char *fmt)
{
	struct tm lt;

	localtime_r(&t, &lt);
	return strftime(outbuf, nbytes, fmt, &lt);
}

static int
scan_field (void *buf, const char *fmt, char **val)
{
	char tmp[256];
	if (1 == sscanf(buf, fmt, tmp)) {
		*val = strdup(tmp);
		return 1;
	}
	return 0;
}

static void
mysql_get_info (void)
{
	char buf[256];
	FILE *fp = fopen(MYSQL_TXT, "r");

	if (!fp)
		err_exit(MYSQL_TXT);
	mysql_hostname = NULL;
	mysql_username = NULL;
	mysql_dbname   = NULL;
	mysql_passwd   = NULL;
	while (fgets(buf, sizeof(buf), fp)) {
		if (!mysql_hostname && scan_field(buf, "DBHostName=%s", &mysql_hostname)) continue;
		if (!mysql_username && scan_field(buf, "DBUserName=%s", &mysql_username)) continue;
		if (!mysql_dbname   && scan_field(buf, "DBName=%s",     &mysql_dbname))   continue;
		if (!mysql_passwd   && scan_field(buf, "DBPassword=%s", &mysql_passwd))   continue;
	}
	fclose(fp);
	if (!mysql_hostname)
		mysql_username = "localhost";
	if (!mysql_username)
		mysql_username = "mythtv";
	if (!mysql_dbname)
		mysql_dbname   = "mythconverg";
	if (!mysql_passwd) {
		syslog(LOG_ERR, "couldn't get mysql passwd\n");
		exit(ENODEV);
	}
}

static int
test_activity (char *buf, int bufsize)
{
	char cmd[] = "/usr/local/bin/set_wakeup_alarm.sh test";
	FILE *fp;
	int n = 0;
	void *rc;

	fp = popen(cmd, "r");
	if (!fp) {
		syslog(LOG_ERR, "popen(set_wakeup_alarm.sh test): %s\n", strerror(errno));
		return 0;
	}
	rc = fgets(buf, bufsize-2, fp);
	fclose(fp);
	if (rc) {
		n = strlen(buf);
		if (n && buf[n-1] == '\n')
			buf[--n] = '\0';
	}
	return n;
}

static int
mysql_get_encoder_list (int *encoders)
{
	/*
	 * The backend crashes if we query a non-existant "remoteencoder" cardid.
	 * So, we need to query the mySQL database first for a full/current list of them.
	 */
	char cmd[1024];
	FILE *fp;
	int n = 0;
	char buf[1024];

	sprintf(cmd, "mysql -h '%s' -u '%s' '--password=%s' '%s'"
		" -e \"SELECT cardid FROM capturecard WHERE hostname='%s' ORDER BY cardid;\"",
		mysql_hostname, mysql_username, mysql_passwd, mysql_dbname, backend_hostname);
	fp = popen(cmd, "r");
	if (!fp) {
		syslog(LOG_ERR, "popen(mysql): %s\n", strerror(errno));
		return 0;
	}
	while (fgets(buf, sizeof(buf), fp)) {
		if (1 == sscanf(buf, "%u", &encoders[n])) {
			//syslog(LOG_INFO, "found encoder %u\n", encoders[n]);
			++n;
		}
	}
	fclose(fp);
	return n;
}

static int
send_backend_request (int sd, const char *text)
{
	int text_len = strlen(text);
	int msg_len = 8 + text_len;
	int rc, err;
	char *msg = malloc(msg_len + 1);

	if (msg == NULL) {
		err = errno;
		syslog(LOG_ERR, "malloc(): %s\n", strerror(err));
		return err;
	}
	sprintf(msg, "%-8u%s", text_len, text);
	rc = write(sd, msg, msg_len);
	err = errno;
	free(msg);
	if (rc != msg_len) {
		if (rc == -1) {
			syslog(LOG_ERR, "write(sd): %s\n", strerror(err));
			return err;
		} else {
			syslog(LOG_ERR, "write(sd,%d): returned %d\n", msg_len, rc);
			return EIO;
		}
	}
	return 0;
}

static char *
get_backend_reply (int sd)
{
	int rc, len, offset = 0;
	char msg_len[9], *msg;

	rc = read(sd, msg_len, 8);
	if (rc < 8) {
		if (rc == -1) {
			syslog(LOG_ERR, "read(socket): %s\n", strerror(errno));
			return NULL;
		}
		syslog(LOG_ERR, "got short reply: %d bytes\n", rc);
		return NULL;
	}
	msg_len[8] = '\0';
	len = atoi(msg_len);
	if (len <= 0) {
		syslog(LOG_ERR, "got short msg_len: %d bytes\n", len);
		return NULL;
	}
	msg = malloc(len + 1);
	if (!msg) {
		syslog(LOG_ERR, "malloc(): %s\n", strerror(errno));
		return NULL;
	}
	do {
		rc = read(sd, msg+offset, len-offset);
		if (rc >= 0) {
			offset += rc;
		} else {
			syslog(LOG_ERR, "read(socket): %s\n", strerror(errno));
			free(msg);
			return NULL;
		}
	} while (offset < len);
	msg[len] = '\0';
	return msg;
}

static char *
do_backend_command (void *vfd, int sd, char *command)
{
	char *reply;

	if (shutdown_sig)
		kill_ourself(vfd);
	if (send_backend_request(sd, command))
		return NULL;
	if (shutdown_sig)
		kill_ourself(vfd);
	reply = get_backend_reply(sd);
	if (shutdown_sig)
		kill_ourself(vfd);
	return reply;
}

static char *
subfield (char *reply, int field, char **internal_p)
{
	int i;
	char *next, *new, saved;
	static const char *magic = "[]:[]";

	if (!reply)
		return NULL;

	for (i = 0; i != field; ++i) {
		if (!*reply || !(next = strstr(reply, magic))) {
			//printf("NULL\n"); fflush(stdout);
			return NULL;
		}
		reply = next + strlen(magic);
	}

	next = strstr(reply, magic);
	if (!next)
		next = reply + strlen(reply);
	saved = *next;
	*next = '\0';
	new = malloc(strlen(reply)+1);
	if (new)
		strcpy(new, reply); 
	*next = saved;
	if (internal_p)
		*internal_p = reply;
	return new;
}

static void
dump_reply_fields (char *reply, int min, int max)
{
	int i;

	for (i = min; i <= max; ++i) {
		char *f = subfield(reply, i, NULL);
		if (f) {
			printf("field[%2u] = \"%s\"\n", i, f);
			free(f);
		}
	}
	putchar('\n');
}

static char *temperature_text = "??\0";

#define GET_TEMPERATURE(FUNCTION,PATH,READ,OPEN,SUFFIX,REFRESH)		\
static char * FUNCTION(time_t now)					\
{									\
	static int failed = 0;						\
	static time_t last = 0;						\
	static char text[8] = "??\0";					\
	const char *path = PATH;					\
	FILE *fp;							\
									\
	if (failed > 3)							\
		return NULL;						\
	if (last && (now - last) < REFRESH)				\
		return text;						\
	last = now;							\
	fp = OPEN(path, "r");						\
	if (fp) {							\
		text[0] = 0;						\
		READ;							\
		fclose(fp);						\
		if (!text[0]) {						\
			/*syslog(LOG_WARNING, "%s: failed1\n", path);*/	\
			++failed;					\
			return NULL;					\
		}							\
		text[2] = SUFFIX;					\
		text[3] = '\0';						\
	} else {							\
		syslog(LOG_WARNING, "%s: %s\n", path, strerror(errno));	\
		++failed;						\
		return NULL;						\
	}								\
	return text;							\
}

// $ nvidia-smi -q -g 0
//
// GPU 0:
// 	Product Name		: GeForce GT 240
// 	PCI ID			: ca310de
// 	Temperature		: 39 C

#if 0
	/* Only run nvidia-smi if nvidia.ko is loaded;
	 * otherwise nvidia-smi tries to load it,
	 * which interferes with driver reinstalls and the like.
	 */
	GET_TEMPERATURE(gpu_temperature,
			"grep -q nvidia /proc/modules && nvidia-smi -q -g 0 || echo ' Temperature : ?? C'",
			{int r; do{r=fscanf(fp,"%s",text); } while (r == 1 && strcmp(text, "Temperature"));if (fscanf(fp," : %s C",text));},
			popen,'G',3)
	#define GPU_TEMPERATURE case 0: text = gpu_temperature(now); break;
#else
	#define GPU_TEMPERATURE
#endif

#if 0
	#define CPU_TEMPERATURE
#else
    #if 1  /* 2.6.37 */
	GET_TEMPERATURE(cpu_temperature,
			"/sys/devices/virtual/thermal/thermal_zone0/temp",
			{	int temp ; \
				if (fscanf(fp, "%d", &temp)) \
					sprintf(text, "%d", (temp + 500) / 1000); \
			},
			fopen,'C',3)
	#define CPU_TEMPERATURE case 1: text = cpu_temperature(now); break;
    #else  /* 2.6.35 */
	GET_TEMPERATURE(cpu_temperature,
			"/proc/acpi/thermal_zone/THRM/temperature",
			if (fscanf(fp, "temperature: %3s C", text));,
			fopen,'C',3)
	#define CPU_TEMPERATURE case 1: text = cpu_temperature(now); break;
    #endif
#endif

#if 0
	GET_TEMPERATURE(dsk_temperature,
			"/usr/sbin/hddtemp -n /dev/sdb",
			if (fgets(text, sizeof(text)-1, fp));,
			popen,'D',3)
	#define DSK_TEMPERATURE case 2: text = dsk_temperature(now); break;
#else
	#define DSK_TEMPERATURE
#endif

static void
get_temperature (time_t now)
{
	static int which = 0;
	static time_t last = 0;
	char *text = NULL;

	if ((now - last) >= 2) {
		last = now;
		++which;
	}

	switch (which) {
		default: which = 0;
		GPU_TEMPERATURE
		CPU_TEMPERATURE
		DSK_TEMPERATURE
	}
	if (text)
		temperature_text = text;
}

static void
format_and_scroll (char *outbuf, const char *text, int tick, const char *temp)
{
	int i, t = 0, start = (tick < 15) ? 15 - tick : 0;
	static int blink = 0;

	if (tick > 15) {
		int tlen = strlen(text);
		t = tick - 15;
		if (t > tlen)
			t = tlen;
	}
	for (i = 0; i < 16; ++i)
		outbuf[i] = (i < start || !text[t]) ? ' ' : text[t++];
	//format_time(outbuf+16, 17, time(NULL), " %H:%M   %d-%b ");
	format_time(outbuf+16, 17, time(NULL), "%H:%M %b-%d ");
	strcat(outbuf, temp);
	if (text[0]) {
		blink = 0;
	} else if (blink++ == 3) {
		outbuf[18] = ' ';
		blink = 0;
	}
}

static unsigned long long
get_msecs (void)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (tv.tv_sec * 1000) + (tv.tv_usec / 1000);
}

static void
show_text (void *vfd, char *text)
{
	char outbuf[33];
	int knob, tick, last_tick = -1, nticks, tlen = strlen(text);

	while (tlen && text[tlen-1] == ' ')
		text[--tlen] = '\0';
	nticks = tlen + ( text[0] ? 16 : 1 );

	for (tick = 0; tick < nticks; ++tick) {
		unsigned long long now = 0, started = get_msecs();
		do {
			if (shutdown_sig)
				kill_ourself(vfd);
			if (tick != last_tick) {
				last_tick = tick;
				if (tick >= 0 && tick < nticks) {
					get_temperature(time(NULL));
					format_and_scroll(outbuf, text, tick, temperature_text);
					outbuf[32] = '\0';
					// This actually writes it to the display:
					if (vfd_write(vfd, outbuf))
						err_exit("vfd_write() failed");
				}
			}
			knob = knob_read(vfd);
			if (knob) {
				if (knob == 1)
					++tick;
				else if (tick > 15)
					--tick;
				started = get_msecs() + 600;
			}
			now = get_msecs();
		} while (now <= started || (now - started) < 150);
	}
}

static char *
rec_scheduled (char *rec, time_t *tstart)
{
	static char recinfo[32];
	char *t;

	/* if recstatus is -1, then rec is NOT scheduled for recording */
	t = subfield(rec, F_recstatus, NULL);
	if (t) {
		if (0 == strcmp(t, "-1")) {
			free(t);
			/*
			 * If rec has been assigned a "card_id" and an "input_id" (on the card),
			 * then rec is scheduled for recording.  Otherwise not.
			 */
			t = subfield(rec, F_cardid,  NULL);
			if (t) {
				char card_id = atoi(t);
				free(t);
				t = subfield(rec, F_inputid, NULL);
				if (t) {
					char input_id = atoi(t);
					free(t);
					if (card_id && input_id)
						recinfo[0] = '\0'; /* normal recording status */
					else if (card_id || input_id)
						snprintf(recinfo, sizeof(recinfo)-1, "<%d.%d?>", card_id, input_id);	/* unexpected */
					t = subfield(rec, F_recstartts, NULL);
					if (t) {
						*tstart = atoi(t);
						free(t);
						return recinfo;
					}
				}
			}
		}
	}
	return NULL;	/* rec is not scheduled for recording */
}

static char *
get_upcoming_recordings (void *vfd, int sd, time_t connect_time, int last_field)
{
	time_t		earliest = 0;
	int		pass, first_match = 0, numrecs = 0, nprograms = 0, outsize = 4096, outx = 0;
	char		*reply, *numrec_s, *out;

	if (!(reply = do_backend_command(vfd, sd, "QUERY_GETALLPENDING")))
		return NULL;
	if (0) dump_reply_fields (reply, 0, 100);	//FIXME

	out = malloc(outsize);
	if (!out)
		return NULL;

	/* get numrecs from start of reply */
	numrec_s = subfield(reply, F_hdr_numrecs, NULL);
	if (!numrec_s) {
		free(out);
		return NULL;
	}
	numrecs = atoi(numrec_s);
	free(numrec_s);

	for (pass = 0; pass < 2; ++pass) {
		int	i, skip = F_first_entry;
		char	*rec = reply;
		/*
		 * pass==0: find earliest recording time from the list.
		 * pass==1: format all recordings which match the earliest time.
		 */
		if (pass == 1) {
			outx += test_activity(out+outx, outsize-outx);
			if (!nprograms) {
				if ((time(NULL) - connect_time) > 5)
					outx += snprintf(out+outx, outsize-outx, "(nothing scheduled?)");
				else
					out[0] = '\0';	/* triggers faster polling by caller */
				break;
			} else {
				time_t now = time(NULL);
				char tstr[32], *fmt = "%d-%b %R";
		
				if (earliest < now || (earliest - now) < (12*3600))
					fmt = "%R";
				format_time(tstr, sizeof(tstr), earliest, fmt);
				outx += snprintf(out+outx, outsize-outx, "Next recording%s at %s: ",
					(nprograms > 1) ? "s" : "", tstr);
			}
		}
		for (i = 0; i < numrecs; ++i) {
			char *next, *scheduled;
			time_t tstart = 0;

			/* point rec at next record */
			next = subfield(rec, skip, &rec);
			if (!next)
				break;
			free(next);
			skip = last_field + 1;

			/* optimization for second pass only: */
			if (i < first_match)
				continue;

			if (0) if (!pass) dump_reply_fields (rec, 0, last_field);	//FIXME
	
			/* get tstart (start time), if scheduled */
			scheduled = rec_scheduled(rec, &tstart);
			if (!scheduled)
				continue;

			if (pass == 0) {
				/* first pass: find earliest recording time */
				if (!earliest || tstart < earliest) {
					earliest = tstart;
					nprograms   = 1;
					first_match = i; /* "i", not "1" */
				} else if (earliest && tstart == earliest)
					++nprograms;
			} else {
				/* second pass: find programs which match earliest recording time */
				if (tstart == earliest) {
					char *channel = subfield(rec, F_chanstr, NULL);
					if (channel) {
						char *title = subfield(rec, F_title, NULL);
						if (title && outx < outsize) {
							outx += snprintf(out+outx, outsize-outx,
									  "Ch.%s%s %s%c  ", channel, scheduled,
									  title, --nprograms ? ',' : '.');
							free(title);
						}
						free(channel);
					}
				}
				if (!nprograms)	/* second pass all done now? */
					break;
			}
		}
	}
	free(reply);
	return out;
}

static char *
get_current_recordings (void *vfd, int sd, int *encoders, int nencoders)
{
	char		command[128];
	unsigned long	max_hours = (7 * 24);
	int		i, outsize = 4096, outx = 0;
	char		*reply, *out = malloc(4096);

	if (!out)
		return NULL;

	// Build up a long string showing everything we are recording right now:
	out[0] = '\0';
	max_hours = (7 * 24);
	for (i = 0; i < nencoders; ++i) {
		static const char *nothing   = "[]:[][]:[][]:[][]:[]0[]:[][]:[][]:";

		sprintf(command, "QUERY_RECORDER %u[]:[]GET_RECORDING", encoders[i]);
		if (!(reply = do_backend_command(vfd, sd, command)))
			return NULL;
/*
 * Hockey Night in Canada: Hockey Tonight[]:[][]:[]The latest news, scores, highlights and game previews with Ron MacLean.[]:[]Sports talk[]:[]2041[]:[]25[]:[]CBOT-DT[]:[]CBC-HD[]:[]/var/lib/mythtv/recordings/2041_20110129183000.mpg[]:[]0[]:[]1296343800[]:[]1296345600[]:[]0[]:[]duke[]:[]2[]:[]24[]:[]2[]:[]25[]:[]-2[]:[]771[]:[]3[]:[]15[]:[]1[]:[]1296343800[]:[]1296345900[]:[]1052672[]:[]Default[]:[][]:[]EP00284160[]:[]SH002841600000[]:[]1296343740[]:[]0[]:[]1998-10-10[]:[]Default[]:[]0[]:[]0[]:[]Default[]:[]1[]:[]1[]:[]1[]:[]0
 */
		if (memcmp(reply, nothing, strlen(nothing))) {
			char *title = subfield(reply, F_title, NULL);
			if (title) {
				char *channel = subfield(reply, F_chanstr, NULL);
				if (channel) {
					char *end = subfield(reply, F_recendts, NULL);
					if (end) {
						outx += snprintf(out+outx, outsize-outx,
							"Ch.%s %s, ends at ", channel, title);
						outx += format_time(out+outx, outsize-outx, atoi(end), "%H:%M.   ");
						free(end);
					}
					free(channel);
				}
				free(title);
			}
		}
		free(reply);
	}
	return out;
}

static int
proto_version_check (void *vfd, int sd, int version)
{
	int	rc = -1;
	char	command[128], *reply, *f1, *f2;
	static	char errmsg[40];

	// verify compatible schema.h version with the backend
	errtext = "Protocol version check..";
	if (version >= 63)
		snprintf(command, sizeof(command), "MYTH_PROTO_VERSION %u %s", version, MYTH_PROTO_TOKEN);
	else
		snprintf(command, sizeof(command), "MYTH_PROTO_VERSION %u", version);
	if (!(reply = do_backend_command(vfd, sd, command)))
		return rc;
	f1 = subfield(reply, 0, NULL);
	f2 = subfield(reply, 1, NULL);
	if (f1 && f2) {
		if (0 == strcmp(f1, "ACCEPT")) {
			rc = 0;
		} else {
			snprintf(errmsg, sizeof(errmsg),
					"Myth protocol mismatch: %u/%s", version, f2);
			errtext = errmsg;
		}
	} else {
		syslog(LOG_ERR, "bad log-in reply: %s\n", reply);
		errtext = "Bad protocol_version reply";
	}
	if (f1) free(f1);
	if (f2) free(f2);
	free(reply);
	return rc;
}

static int
connect_to_backend (void *vfd, int version)
{
	int sd, rc;
	struct sockaddr_in saddr;

	// create a new socket object:
	sd = socket(PF_INET, SOCK_STREAM, 6 /* TCP */);
	if (sd == -1)
		err_exit("socket() failed");

	// establish a "name" (ipaddr:port) for the server:
	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family = AF_INET;
	saddr.sin_addr.s_addr = htonl(0x7f000001);// localhost: 127.0.0.1
	saddr.sin_port = htons(6543);		// mythbackend port number

	// (re-)connect to the server:
	rc = connect(sd, (struct sockaddr *)&saddr, sizeof(saddr));
	if (rc == 0) {
		rc = proto_version_check(vfd, sd, version);
		if (rc)
			errno = -1;
	}
	if (rc) {
		close(sd);
		sd = -1;
	}
	return sd;
}

static int
monitor_backend (void *vfd)
{
	int	sd, encoders[32], nencoders, last_field;
	char	*reply, command[128];
	time_t	lasttime = 0, connect_time;

	if ((sd = connect_to_backend(vfd, MYTH_PROTO_VERSION)) != -1)
		last_field = P63_last_field;
#if 0
	else if ((sd = connect_to_backend(vfd, 23056)) != -1)
		last_field = P50_last_field;
	else if ((sd = connect_to_backend(vfd, 56)) != -1)
		last_field = P50_last_field;
	else if ((sd = connect_to_backend(vfd, 50)) != -1)
		last_field = P40_last_field;
	else if ((sd = connect_to_backend(vfd, 40)) != -1)
		last_field = P40_last_field;
	else if ((sd = connect_to_backend(vfd, 31)) != -1)
		last_field = P31_last_field;
#endif
	else
		return errno;
	
	// Get a list of recording devices from mySQL
	// for later use in generating queries to the backend:
	nencoders = mysql_get_encoder_list(encoders);

	// Register with the backend as a "Monitor" machine:
	errtext = "Registering..";
	sprintf(command, "ANN Monitor %s 0\n", backend_hostname);
	if (!(reply = do_backend_command(vfd, sd, command)))
		return -1;
	free(reply);
	connect_time = time(NULL);

	// Now just loop, querying the backend and showing the results:
	for (;;) {
		int is_recording;
		char *reply, *text;

		errtext = "Querying state..";
		if (!(reply = do_backend_command(vfd, sd, "QUERY_ISRECORDING")))
			break;
		is_recording = (reply[0] != '0');
		free(reply);
	
		text = NULL;
		if (is_recording) {
			errtext = "Querying recordings..";
			text = get_current_recordings(vfd, sd, encoders, nencoders);
		} else if (!lasttime || (time(NULL) - lasttime) >= 10) {
			errtext = "Querying schedule..";
			text = get_upcoming_recordings(vfd, sd, connect_time, last_field);
		} else {
			show_text(vfd, "");
			usleep(250000);
			continue;
		}
		if (!text)
			break;
		show_text(vfd, text);
		if (text[0]) {
			lasttime = time(NULL);
		} else {
			lasttime = 0;
			usleep(250000);
		}
		free(text);
	}
	return -1;
}

static char *
get_basename (char *pathname)
{
	char *basename = strrchr(pathname, '/');
	if (!basename || !*++basename)
		basename = pathname;
	return basename;
}

int
main (int argc, char *argv[])
{
	void *vfd = NULL;
	char *myname;
	time_t started = time(NULL);

	myname = get_basename(argv[0]);
	mysql_get_info();
	// if requested, detach from the tty and run as a background daemon
	if (argc > 1 && 0 == strcmp(argv[1], "-d")) {
		if (-1 == daemon(0,0))
			err_exit("daemon() failed");
	} else {
		if (chdir("/"))
			err_exit("chdir(\"/\") failed");
	}
	openlog(myname, LOG_CONS|LOG_PID|LOG_PERROR, LOG_DAEMON);

	/* we use our own hostname as the "backend" hostname in a couple of places: */
	gethostname(backend_hostname, sizeof(backend_hostname));
	mlockall(MCL_CURRENT | MCL_FUTURE);

try_open:
	vfd = vfd_open();
	if (!vfd) {
		//syslog(LOG_ERR, "retrying in 5 seconds\n");	//FIXME ?
		sleep(5);
		goto try_open;
	}

	// We trap common signals, to allow blanking the display
	// before our program exits.  This way, the user can see
	// that we're not actually updating the display anymore.
	//
	signal(SIGINT,  my_sighandler);
	signal(SIGTERM, my_sighandler);
	signal(SIGQUIT, my_sighandler);
	signal(SIGABRT, my_sighandler);
	signal(SIGPIPE, my_sighandler);

	show_text(vfd, "");
	while (1) {
		int err;
		static int last_err;
		errtext = "Protocol-error.";
		err = monitor_backend(vfd);
		if (err != -1) {
			if (err != ECONNREFUSED && err != ETIMEDOUT && err != EAGAIN) {
				syslog(LOG_ERR, "err=%d\n", err);	//FIXME
				err_exit("connect() failed");
			}
			if (err != last_err) {
				syslog(LOG_ERR, "connect(): %s\n", strerror(err));
				last_err = err;
			}
			sleep(2);
			if ((time(NULL) - started) > 20)
				errtext = "MythTV not responding.";
			else
				errtext = "Connecting..";
		}
		if (shutdown_sig)
			kill_ourself(vfd);
		show_text(vfd, errtext);
		if (shutdown_sig)
			kill_ourself(vfd);
	}
	vfd_close(vfd);
	return 0;
}
