#include <stdio.h>
#include <usb.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>

#include <sys/time.h>
#include <time.h>

static struct usb_device *
scan_for_usb_dev (struct usb_bus *busses, unsigned int idVendor, unsigned int idProduct)
{
	struct usb_bus *bus;

	for (bus = busses; bus; bus = bus->next) {
		struct usb_device *dev;

		for (dev = bus->devices; dev; dev = dev->next) {
			if (dev->descriptor.idVendor == idVendor && dev->descriptor.idProduct) {
				return dev;
			}
		}
	}
	return NULL;
}

static int
release_all_interfaces (struct usb_device *dev, usb_dev_handle *dh)
{
	int c, i, a;

	for (c = 0; c < dev->descriptor.bNumConfigurations; ++c) {
		struct usb_config_descriptor *config = &dev->config[c];
		for (i = 0; i < config->bNumInterfaces; ++i) {
			struct usb_interface *interface = &config->interface[i];
			for (a = 0; a < interface->num_altsetting; ++a) {
				struct usb_interface_descriptor *iface = &interface->altsetting[a];
				int inum =  iface->bInterfaceNumber;
				int rc = usb_release_interface(dh, inum);
				if (rc) {
					syslog(LOG_ERR, "usb_release_interface(%d) failed, rc=%d (%s)\n",
						       inum, rc, strerror(-rc));
					return rc;
				}
			}
		}
	}
	return 0;
}

static int
claim_all_interfaces (struct usb_device *dev, usb_dev_handle *dh)
{
	int c, i, a;

	for (c = 0; c < dev->descriptor.bNumConfigurations; ++c) {
		struct usb_config_descriptor *config = &dev->config[c];
		for (i = 0; i < config->bNumInterfaces; ++i) {
			struct usb_interface *interface = &config->interface[i];
			for (a = 0; a < interface->num_altsetting; ++a) {
				struct usb_interface_descriptor *iface = &interface->altsetting[a];
				char name[32];
				int inum =  iface->bInterfaceNumber;
				int rc = usb_get_driver_np(dh, inum, name, sizeof(name));
				if (rc == 0) {
					if (0 == strcmp("usbfs", name)) {
						syslog(LOG_ERR, "device already claimed (busy) by usbfs\n");
						return -EBUSY;
					}
					syslog(LOG_INFO, "detaching inum %d from kernel driver \"%s\"\n", inum, name);
					rc = usb_detach_kernel_driver_np(dh, inum);
					if (rc) {
						syslog(LOG_ERR, "usb_detach_kernel_driver_np(%d) failed, rc=%d (%s)\n",
						       	inum, rc, strerror(-rc));
						return rc;
					}
				} else if (rc != -ENODATA) {
					syslog(LOG_ERR, "usb_get_driver_np(%d) failed, rc=%d (%s)\n",
						       inum, rc, strerror(-rc));
					return rc;
				}
				rc = usb_claim_interface(dh, inum); 
				if (rc) {
					syslog(LOG_ERR, "usb_claim_interface(%d) failed, rc=%d (%s)\n",
						       inum, rc, strerror(-rc));
					return rc;
				}
			}
		}
	}
	return 0;
}

static struct usb_bus		*busses = NULL;
static int			extra_packet_required = 0;
static struct usb_device	*vfd_dev = NULL;
static int			vfd_endpoint = 0;
static int			knob_endpoint = 0;

/*
 * VFD information:
 *
 * one interface.
 * 	the first "input"  endpoint is for the IR receiver (or knob).
 * 	the first "output" endpoint is for writing to the VFD (display).
 *
 * The display is 2 rows of 16 characters each,
 * sent as four packets of 7 data bytes, and one packet of 4 data bytes.
 * Pad the excess packet bytes with 0xff.
 * If vendor is 0x15c2, a 6th packet is required (???).
 *
 * Fill usb_tx_buf[0..6] with data, and usb_tx_buf[7] with a sequence number {0..4}.
 * For the optional 6th packet, fill [0..6] with 1,0,0,0,0,ff,ff, and [7] with seqno {5}.
 *
 * These packets are just ordinary int_urb's.
 * Send them with usb_interrupt_write(dh, endpoint, *bytes, bytecount, timeout);
 * This returns a count of actual bytes written, or -errno.
 */

int
vfd_write (usb_dev_handle *dh, const char *msg)
{
	int seqno, i, rc;
	unsigned char data[9];

	for (seqno = 0; seqno <= (4 + extra_packet_required); ++seqno) {
		if (seqno == 5) {
			data[0] = 0x01;
			data[1] = data[2] = data[3] = data[4] = 0x00;
			data[5] = data[6] = 0xff;
		} else {
			for (i = 0; i < 7; ++i)
				data[i] = *msg ? *msg++ : ' ';
			if (seqno == 4)
				data[4] = data[5] = data[6] = 0xff;
		}
		data[7] = seqno << 1;
		rc = usb_interrupt_write(dh, vfd_endpoint, (void*)data, 8, 1000 /*msec*/);
		if (rc != 8) {
			syslog(LOG_ERR, "usb_interrupt_write() returned %d instead of %d\n", rc, 8);
			return (rc < 0) ? rc : -EINVAL;
		}
	}
	return 0;
}

/*
 * By default, reading from the knob endpoint always returns data in 7msec or so.
 * Even if there's nothing (new) to report.  We probably should be setting some
 * control register or something first, but ??? no documentation.
 * So, just poll, and poll.. and poll..
 *
 * Also, every second packet seems to be all 0xff regardless,
 * so it might be useful for an app to call this in pairs,
 * if the first result it sees is 0.
 *
 * This function returns 0 for no movement, 1 for LEFT, or 2 for RIGHT.
 */
int
knob_read (usb_dev_handle *dh)
{
	int rc;
	unsigned char data[8];

	rc = usb_interrupt_read(dh, knob_endpoint, (void*)data, 8, 1000 /*millisecs, ignored*/);
	//syslog(LOG_ERR, "rc=%d, d0=0x%02x d1=0x%02x d2=0x%02x\n", rc, data[0], data[1], data[2]);
	if (rc == 8 && data[2] != 0xff) {
		if (data[0])
			return 1;
		if (data[1])
			return 2;
	}
	return 0;
}

static int
get_vfd_and_knob_endpoints (struct usb_device *dev, int *vfd_endpoint, int *knob_endpoint)
{
	int c, i, a, e, got_vfd = 0, got_knob = 0;

	for (c = 0; c < dev->descriptor.bNumConfigurations; ++c) {
		struct usb_config_descriptor *config = &dev->config[c];
		for (i = 0; i < config->bNumInterfaces; ++i) {
			struct usb_interface *interface = &config->interface[i];
			for (a = 0; a < interface->num_altsetting; ++a) {
				struct usb_interface_descriptor *iface = &interface->altsetting[a];
				for (e = 0; e < iface->bNumEndpoints; ++e) {
					struct usb_endpoint_descriptor *ep = &iface->endpoint[e];
					if ((ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_ENDPOINT_IN) {
						if (!got_knob++)
							*knob_endpoint = ep->bEndpointAddress;
					} else if ((ep->bmAttributes & USB_ENDPOINT_TYPE_MASK) == USB_ENDPOINT_TYPE_INTERRUPT) {
						if (!got_vfd++)
							*vfd_endpoint = ep->bEndpointAddress;
					}
				}
			}
		}
	}
	if (got_knob && got_vfd)
		return 0;
	syslog(LOG_ERR, "get_vfd_and_knob_endpoints:%s%s\n",
		got_vfd ? "" : " VFD endpoint not found", got_knob  ? "" : " KNOB endpoint not found");
	return EINVAL;
}

static void *
do_perror (const char *msg, int err)
{
	errno = err;
	//perror(msg);
	syslog(LOG_ERR, "%s: %s\n", msg, strerror(err));
	errno = err;
	return NULL;
}

void vfd_close (void *dh)
{
	release_all_interfaces(vfd_dev, dh);
	usb_close(dh);
}

void *vfd_open (void)
{
	usb_dev_handle *dh;
	int rc;

	if (!vfd_dev) {
		usb_init();
		usb_find_busses();
		usb_find_devices();
		busses = usb_get_busses();
		if (!vfd_dev) vfd_dev = scan_for_usb_dev(busses, 0x0aa8, 0xffda);
		if (!vfd_dev) vfd_dev = scan_for_usb_dev(busses, 0x15c2, 0xffda);
		if (!vfd_dev) vfd_dev = scan_for_usb_dev(busses, 0x15c2, 0xffdc);
		if (!vfd_dev) vfd_dev = scan_for_usb_dev(busses, 0x15c9, 0x0038);
		if (!vfd_dev)
			return do_perror("VFD not found", ENODEV);
	}
	dh = usb_open(vfd_dev);
	if (dh == NULL)
		return do_perror("usb_open", ENODEV);
	rc = get_vfd_and_knob_endpoints(vfd_dev, &vfd_endpoint, &knob_endpoint); 
	if (rc) {
		usb_close(dh);
		return do_perror("get_vfd_and_knob_endpoints", rc);
	}
	extra_packet_required = (vfd_dev->descriptor.idVendor == 0x15c2);

	rc = claim_all_interfaces(vfd_dev, dh);
	if (rc) {
		usb_close(dh);
		errno = -rc;
		return NULL;
	}

	rc = usb_clear_halt(dh, vfd_endpoint);
	if (rc < 0) {
		vfd_close(dh);
		return do_perror("usb_clear_halt", -rc);
	}

	rc = vfd_write(dh, "");
	if (rc < 0) {
		vfd_close(dh);
		errno = -rc;
		return NULL;
	}
	return dh;
}
